local _, MazeHelper = ...;
MazeHelper.E = {};

local E = MazeHelper.E;
local M = MazeHelper.M;

E.CreateRoundedCheckButton = function(parent)
    local b = CreateFrame('CheckButton', nil, parent);

    b:SetNormalTexture(M.Icons.TEXTURE);
    b:GetNormalTexture():SetTexCoord(unpack(M.Icons.COORDS.CHECKBUTTON_NORMAL));
    b:SetCheckedTexture(M.Icons.TEXTURE);
    b:GetCheckedTexture():SetTexCoord(unpack(M.Icons.COORDS.CHECKBUTTON_CHECKED));
    b:SetHighlightTexture(M.Icons.TEXTURE);
    b:GetHighlightTexture():SetTexCoord(unpack(M.Icons.COORDS.CHECKBUTTON_NORMAL));

    b.Label = b:CreateFontString(nil, 'ARTWORK', 'GameFontNormal');
    PixelUtil.SetPoint(b.Label, 'LEFT', b, 'RIGHT', 6, 0);

    b.SetLabel = function(self, label)
        self.Label:SetText(label);
        self:SetHitRectInsets(0, -1 * math.max(100, self.Label:GetStringWidth() + 4), 0, 0);
    end

    b.SetTooltip = function(self, tooltip)
        self.tooltip = tooltip;
    end

    b.SetPosition = function(self, point, relativeTo, relativePoint, offsetX, offsetY, minOffsetXPixels, minOffsetYPixels)
        PixelUtil.SetPoint(self, point, relativeTo, relativePoint, offsetX, offsetY, minOffsetXPixels, minOffsetYPixels);
    end

    -- Area - because i don't want to touch the original SetSize and it's just a synonym for Size and nothing else
    b.SetArea = function(self, width, height)
        PixelUtil.SetSize(self, width, height);
    end

    b:HookScript('OnEnter', function(self)
        if not self.tooltip then
            return;
        end

        GameTooltip:SetOwner(self, 'ANCHOR_RIGHT');
        GameTooltip:AddLine(self.tooltip, 1, 0.85, 0, true);
        GameTooltip:Show();
    end);

    b:HookScript('OnLeave', GameTooltip_Hide);

    return b;
end

E.CreateCheckButton = function(name, parent)
    local b = CreateFrame('CheckButton', name, parent, 'ChatConfigCheckButtonTemplate');

    b.Label = _G[b:GetName() .. 'Text'];
    PixelUtil.SetPoint(b.Label, 'LEFT', b, 'RIGHT', 4, 0);

    b.SetLabel = function(self, label)
        self.Label:SetText(label);
        self:SetHitRectInsets(0, -(self.Label:GetWidth() + 8), 0, 0);
    end

    b.SetPosition = function(self, point, relativeTo, relativePoint, offsetX, offsetY)
        PixelUtil.SetPoint(self, point, relativeTo, relativePoint, offsetX, offsetY);
    end

    b.SetTooltip = function(self, tooltip)
        self.tooltip = tooltip;
    end

    b:HookScript('OnEnter', function(self)
        if not self.tooltip then
            return;
        end

        GameTooltip:SetOwner(self, 'ANCHOR_RIGHT');
        GameTooltip:AddLine(self.tooltip, 1, 0.85, 0, true);
        GameTooltip:Show();
    end);

    b:HookScript('OnLeave', GameTooltip_Hide);

    hooksecurefunc(b, 'SetEnabled', function(self, state)
        if state then
            self.Label:SetTextColor(1, 1, 1);
        else
            self.Label:SetTextColor(0.35, 0.35, 0.35);
        end
    end);

    hooksecurefunc(b, 'Enable', function(self)
        self.Label:SetTextColor(1, 1, 1);
    end);

    hooksecurefunc(b, 'Disable', function(self)
        self.Label:SetTextColor(0.35, 0.35, 0.35);
    end);

    return b;
end