local ADDON_NAME, MazeHelper = ...;
local L = MazeHelper.L;
local E = MazeHelper.E;
local M = MazeHelper.M;

-- Lua API
local tonumber = tonumber;

-- WoW API
local IsInRaid, IsInGroup, UnitIsGroupLeader, GetMinimapZoneText = IsInRaid, IsInGroup, UnitIsGroupLeader, GetMinimapZoneText;

C_ChatInfo.RegisterAddonMessagePrefix('MAZEHELPER');
C_ChatInfo.RegisterAddonMessagePrefix('TIRNASCYTHE');

local playerNameWithRealm, playerRole, inInstance, bossKilled, inEncounter, isMinimized;
local startedInMinMode = false;

local FRAME_SIZE = 300;
local X_OFFSET = 2;
local Y_OFFSET = -2;

local MAX_BUTTONS = 8;
local MAX_ACTIVE_BUTTONS = 4;
local NUM_ACTIVE_BUTTONS = 0;

local PASSED_COUNTER = 1;
local SOLUTION_BUTTON_ID;

local MISTCALLER_ENCOUNTER_ID = 2392;

local buttons = {}
local buttonsData = {
    [1] = {
        name = L['MAZE_HELPER_LEAF_FULL_CIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_LEAF_FULL_CIRCLE'],
        coords = M.Symbols.COORDS_COLOR.LEAF_CIRCLE_FILL,
        coords_white = M.Symbols.COORDS_WHITE.LEAF_CIRCLE_FILL,
        leaf = true,
        flower = false,
        circle = true,
        fill = true,
    },
    [2] = {
        name = L['MAZE_HELPER_LEAF_NOFULL_CIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_LEAF_NOFULL_CIRCLE'],
        coords = M.Symbols.COORDS_COLOR.LEAF_CIRCLE_NOFILL,
        coords_white = M.Symbols.COORDS_WHITE.LEAF_CIRCLE_NOFILL,
        leaf = true,
        flower = false,
        circle = true,
        fill = false,
    },
    [3] = {
        name = L['MAZE_HELPER_FLOWER_FULL_CIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_FLOWER_FULL_CIRCLE'],
        coords = M.Symbols.COORDS_COLOR.FLOWER_CIRCLE_FILL,
        coords_white = M.Symbols.COORDS_WHITE.FLOWER_CIRCLE_FILL,
        leaf = false,
        flower = true,
        circle = true,
        fill = true,
    },
    [4] = {
        name = L['MAZE_HELPER_FLOWER_NOFULL_CIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_FLOWER_NOFULL_CIRCLE'],
        coords = M.Symbols.COORDS_COLOR.FLOWER_CIRCLE_NOFILL,
        coords_white = M.Symbols.COORDS_WHITE.FLOWER_CIRCLE_NOFILL,
        leaf = false,
        flower = true,
        circle = true,
        fill = false,
    },
    [5] = {
        name = L['MAZE_HELPER_LEAF_FULL_NOCIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_LEAF_FULL_NOCIRCLE'],
        coords = M.Symbols.COORDS_COLOR.LEAF_NOCIRCLE_FILL,
        coords_white = M.Symbols.COORDS_WHITE.LEAF_NOCIRCLE_FILL,
        leaf = true,
        flower = false,
        circle = false,
        fill = true,
    },
    [6] = {
        name = L['MAZE_HELPER_LEAF_NOFULL_NOCIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_LEAF_NOFULL_NOCIRCLE'],
        coords = M.Symbols.COORDS_COLOR.LEAF_NOCIRCLE_NOFILL,
        coords_white = M.Symbols.COORDS_WHITE.LEAF_NOCIRCLE_NOFILL,
        leaf = true,
        flower = false,
        circle = false,
        fill = false,
    },
    [7] = {
        name = L['MAZE_HELPER_FLOWER_FULL_NOCIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_FLOWER_FULL_NOCIRCLE'],
        coords = M.Symbols.COORDS_COLOR.FLOWER_NOCIRCLE_FILL,
        coords_white = M.Symbols.COORDS_WHITE.FLOWER_NOCIRCLE_FILL,
        leaf = false,
        flower = true,
        circle = false,
        fill = true,
    },
    [8] = {
        name = L['MAZE_HELPER_FLOWER_NOFULL_NOCIRCLE'],
        aname = L['MAZE_HELPER_ANNOUNCE_FLOWER_NOFULL_NOCIRCLE'],
        coords = M.Symbols.COORDS_COLOR.FLOWER_NOCIRCLE_NOFILL,
        coords_white = M.Symbols.COORDS_WHITE.FLOWER_NOCIRCLE_NOFILL,
        leaf = false,
        flower = true,
        circle = false,
        fill = false,
    },
};

local buttonsMOTSHMap = {
    [1] = 4,
    [2] = 3,
    [3] = 2,
    [4] = 1,
    [5] = 8,
    [6] = 7,
    [7] = 6,
    [8] = 5,
};

local buttonsMYtoWA = {
    [1] = '0 1 0 0',
    [2] = '0 1 1 1',
    [3] = '0 0 0 2',
    [4] = '0 0 1 3',
    [5] = '1 1 0 4',
    [6] = '1 1 1 5',
    [7] = '1 0 0 6',
    [8] = '1 0 1 7',
}

local buttonsWAtoMY = {
    ['0 1 0 0'] = 1,
    ['0 1 1 1'] = 2,
    ['0 0 0 2'] = 3,
    ['0 0 1 3'] = 4,
    ['1 1 0 4'] = 5,
    ['1 1 1 5'] = 6,
    ['1 0 0 6'] = 7,
    ['1 0 1 7'] = 8,
};

local function GetPartyChatType()
    return (not IsInRaid() and IsInGroup(LE_PARTY_CATEGORY_INSTANCE)) and 'INSTANCE_CHAT' or (IsInGroup(LE_PARTY_CATEGORY_HOME) and 'PARTY' or false);
end

local function BetterOnDragStop(frame)
    local point, relativeTo, relativePoint, xOfs, yOfs = frame:GetPoint();

    frame:StopMovingOrSizing();

    frame:ClearAllPoints();
    PixelUtil.SetPoint(frame, point, relativeTo, relativePoint, xOfs, yOfs);
    frame:SetUserPlaced(true);
end

local function SwitchSymbolsColorMode(colored)
    for i = 1, MAX_BUTTONS do
        buttons[i].Icon:SetTexCoord(unpack(colored and buttonsData[i].coords or buttonsData[i].coords_white));
    end
end

MazeHelper.frame = CreateFrame('Frame', 'ST_Maze_Helper', UIParent);
PixelUtil.SetPoint(MazeHelper.frame, 'CENTER', UIParent, 'CENTER', -FRAME_SIZE, FRAME_SIZE);
PixelUtil.SetSize(MazeHelper.frame, FRAME_SIZE + X_OFFSET * (MAX_ACTIVE_BUTTONS - 1), FRAME_SIZE * 3/4);
MazeHelper.frame:EnableMouse(true);
MazeHelper.frame:SetMovable(true);
MazeHelper.frame:SetClampedToScreen(true);
MazeHelper.frame:RegisterForDrag('LeftButton');
MazeHelper.frame:SetScript('OnDragStart', function(self)
    if self:IsMovable() then
        self:StartMoving();
    end
end);
MazeHelper.frame:SetScript('OnDragStop', BetterOnDragStop);

-- Background
MazeHelper.frame.background = MazeHelper.frame:CreateTexture(nil, 'BACKGROUND');
PixelUtil.SetPoint(MazeHelper.frame.background, 'TOPLEFT', MazeHelper.frame, 'TOPLEFT', -15, 22);
PixelUtil.SetPoint(MazeHelper.frame.background, 'BOTTOMRIGHT', MazeHelper.frame, 'BOTTOMRIGHT', 15, -98);
MazeHelper.frame.background:SetTexture(M.BACKGROUND_WHITE);
MazeHelper.frame.background:SetVertexColor(0.05, 0.05, 0.05);
MazeHelper.frame.background:SetAlpha(0.85);

-- Close Button
MazeHelper.frame.CloseButton = CreateFrame('Button', nil, MazeHelper.frame);
PixelUtil.SetPoint(MazeHelper.frame.CloseButton, 'TOPRIGHT', MazeHelper.frame, 'TOPRIGHT', -8, -4);
PixelUtil.SetSize(MazeHelper.frame.CloseButton, 10, 10);
MazeHelper.frame.CloseButton:SetNormalTexture(M.Icons.TEXTURE);
MazeHelper.frame.CloseButton:GetNormalTexture():SetTexCoord(unpack(M.Icons.COORDS.CROSS_WHITE));
MazeHelper.frame.CloseButton:GetNormalTexture():SetVertexColor(0.7, 0.7, 0.7, 1);
MazeHelper.frame.CloseButton:SetHighlightTexture(M.Icons.TEXTURE, 'BLEND');
MazeHelper.frame.CloseButton:GetHighlightTexture():SetTexCoord(unpack(M.Icons.COORDS.CROSS_WHITE));
MazeHelper.frame.CloseButton:GetHighlightTexture():SetVertexColor(1, 0.85, 0, 1);
MazeHelper.frame.CloseButton:SetScript('OnClick', function(self)
    if MazeHelper.frame.Settings:IsShown() then
        MazeHelper.frame.SettingsButton:Click();
    end

    MazeHelper.frame:SetShown(false);
end);

-- Settings Button
MazeHelper.frame.SettingsButton = CreateFrame('Button', nil, MazeHelper.frame);
PixelUtil.SetPoint(MazeHelper.frame.SettingsButton, 'RIGHT', MazeHelper.frame.CloseButton, 'LEFT', -10, 0);
PixelUtil.SetSize(MazeHelper.frame.SettingsButton, 11, 11);
MazeHelper.frame.SettingsButton:SetNormalTexture(M.Icons.TEXTURE);
MazeHelper.frame.SettingsButton:GetNormalTexture():SetTexCoord(unpack(M.Icons.COORDS.GEAR_WHITE));
MazeHelper.frame.SettingsButton:GetNormalTexture():SetVertexColor(0.7, 0.7, 0.7, 1);
MazeHelper.frame.SettingsButton:SetHighlightTexture(M.Icons.TEXTURE, 'BLEND');
MazeHelper.frame.SettingsButton:GetHighlightTexture():SetTexCoord(unpack(M.Icons.COORDS.GEAR_WHITE));
MazeHelper.frame.SettingsButton:GetHighlightTexture():SetVertexColor(1, 0.85, 0, 1);
MazeHelper.frame.SettingsButton:SetScript('OnClick', function(self)
    local settingsIsShown = MazeHelper.frame.Settings:IsShown();

    if not settingsIsShown then
        self:LockHighlight();
    else
        self:UnlockHighlight();
    end

    MazeHelper.frame.Settings:SetShown(not settingsIsShown);
    MazeHelper.frame.MainHolder:SetShown(settingsIsShown);
    MazeHelper.frame.MinButton:SetShown(settingsIsShown);
end);

-- Minimize Button
MazeHelper.frame.MinButton = CreateFrame('Button', nil, MazeHelper.frame);
PixelUtil.SetPoint(MazeHelper.frame.MinButton, 'RIGHT', MazeHelper.frame.SettingsButton, 'LEFT', -8, 0);
PixelUtil.SetSize(MazeHelper.frame.MinButton, 14, 14);
MazeHelper.frame.MinButton.icon = MazeHelper.frame.MinButton:CreateTexture(nil, 'OVERLAY');
PixelUtil.SetPoint(MazeHelper.frame.MinButton.icon, 'BOTTOM', MazeHelper.frame.MinButton, 'BOTTOM', 0, 2);
PixelUtil.SetSize(MazeHelper.frame.MinButton.icon, 10, 2);
MazeHelper.frame.MinButton.icon:SetTexture('Interface\\Buttons\\WHITE8x8');
MazeHelper.frame.MinButton.icon:SetVertexColor(0.7, 0.7, 0.7, 1);
MazeHelper.frame.MinButton:SetScript('OnClick', function(self)
    isMinimized = true;

    PixelUtil.SetHeight(MazeHelper.frame, 40);
    for i = 1, MAX_BUTTONS do
        buttons[i]:Hide();
    end

    PixelUtil.SetPoint(MazeHelper.frame.SolutionText, 'LEFT', MazeHelper.frame, 'LEFT', 2, 4);

    MazeHelper.frame.BottomButtonsHolder:SetShown(false);

    PixelUtil.SetPoint(MazeHelper.frame.background, 'TOPLEFT', MazeHelper.frame, 'TOPLEFT', -15, 0);
    PixelUtil.SetPoint(MazeHelper.frame.background, 'BOTTOMRIGHT', MazeHelper.frame, 'BOTTOMRIGHT', 15, 0);

    PixelUtil.SetPoint(MazeHelper.frame.CloseButton, 'TOPRIGHT', MazeHelper.frame, 'TOPRIGHT', -8, -9);

    MazeHelper.frame.PassedCounter:ClearAllPoints();
    PixelUtil.SetPoint(MazeHelper.frame.PassedCounter, 'LEFT', MazeHelper.frame, 'LEFT', -18, 5);
    MazeHelper.frame.PassedCounter:SetScale(1);
    PixelUtil.SetPoint(MazeHelper.frame.PassedCounter.Text, 'CENTER', MazeHelper.frame.PassedCounter, 'CENTER', (PASSED_COUNTER == 1) and -2 or 0, 0);

    MazeHelper.frame.MiniSolution:SetShown(SOLUTION_BUTTON_ID and true or false);
    if SOLUTION_BUTTON_ID then
        MazeHelper.frame.MiniSolution.Icon:SetTexCoord(unpack(MHMOTSConfig.UseColoredSymbols and buttonsData[SOLUTION_BUTTON_ID].coords or buttonsData[SOLUTION_BUTTON_ID].coords_white));
    end

    MazeHelper.frame.PassedCounter:SetShown(not MazeHelper.frame.MiniSolution:IsShown());

    MazeHelper.frame.AnnounceButton:SetShown(false);

    MazeHelper.frame.SettingsButton:SetShown(false);

    MazeHelper.frame.InvisibleMaxButton:SetShown(true);

    MazeHelper.frame.MinButton:SetShown(false);
end);
MazeHelper.frame.MinButton:SetScript('OnEnter', function(self) self.icon:SetVertexColor(1, 0.85, 0, 1); end);
MazeHelper.frame.MinButton:SetScript('OnLeave', function(self) self.icon:SetVertexColor(0.7, 0.7, 0.7, 1); end);

-- Invisible Maximize Button
MazeHelper.frame.InvisibleMaxButton = CreateFrame('Button', nil, MazeHelper.frame);
PixelUtil.SetPoint(MazeHelper.frame.InvisibleMaxButton, 'TOPLEFT', MazeHelper.frame, 'TOPLEFT', 0, 0);
PixelUtil.SetPoint(MazeHelper.frame.InvisibleMaxButton, 'BOTTOMRIGHT', MazeHelper.frame, 'BOTTOMRIGHT', 0, 0);
MazeHelper.frame.InvisibleMaxButton:SetScript('OnClick', function()
    if not isMinimized then
        return;
    end

    isMinimized = false;

    PixelUtil.SetHeight(MazeHelper.frame, FRAME_SIZE * 3/4);
    for i = 1, MAX_BUTTONS do
        buttons[i]:Show();
    end

    PixelUtil.SetPoint(MazeHelper.frame.SolutionText, 'LEFT', MazeHelper.frame, 'LEFT', 2, -54);

    MazeHelper.frame.BottomButtonsHolder:SetShown(true);
    MazeHelper.frame.PassedButton:SetShown(not inEncounter);
    if inEncounter then
        PixelUtil.SetSize(MazeHelper.frame.BottomButtonsHolder, MazeHelper.frame.ResetButton:GetWidth(), 22);
    else
        PixelUtil.SetSize(MazeHelper.frame.BottomButtonsHolder, MazeHelper.frame.ResetButton:GetWidth() + MazeHelper.frame.PassedButton:GetWidth() + 8, 22);
    end

    PixelUtil.SetPoint(MazeHelper.frame.background, 'TOPLEFT', MazeHelper.frame, 'TOPLEFT', -15, 22);
    PixelUtil.SetPoint(MazeHelper.frame.background, 'BOTTOMRIGHT', MazeHelper.frame, 'BOTTOMRIGHT', 15, -98);

    PixelUtil.SetPoint(MazeHelper.frame.CloseButton, 'TOPRIGHT', MazeHelper.frame, 'TOPRIGHT', -8, -4);

    MazeHelper.frame.PassedCounter:ClearAllPoints();
    PixelUtil.SetPoint(MazeHelper.frame.PassedCounter, 'BOTTOM', MazeHelper.frame, 'TOP', 0, -32);
    MazeHelper.frame.PassedCounter:SetScale(1.25);
    PixelUtil.SetPoint(MazeHelper.frame.PassedCounter.Text, 'CENTER', MazeHelper.frame.PassedCounter, 'CENTER', (PASSED_COUNTER == 1) and -2 or 0, -1);
    MazeHelper.frame.PassedCounter:SetShown(true);

    MazeHelper.frame.MiniSolution:SetShown(false);

    MazeHelper.frame.AnnounceButton:SetShown((MazeHelper.frame.AnnounceButton.solutionAnnounceName and GetPartyChatType() and not MHMOTSConfig.AutoAnnouncer) and true or false);

    MazeHelper.frame.SettingsButton:SetShown(true);

    MazeHelper.frame.InvisibleMaxButton:SetShown(false);

    MazeHelper.frame.MinButton:Show();
end);
MazeHelper.frame.InvisibleMaxButton:RegisterForDrag('LeftButton');
MazeHelper.frame.InvisibleMaxButton:SetScript('OnDragStart', function()
    if MazeHelper.frame:IsMovable() then
        MazeHelper.frame:StartMoving();
    end
end);
MazeHelper.frame.InvisibleMaxButton:SetScript('OnDragStop', function()
    BetterOnDragStop(MazeHelper.frame);
end);
MazeHelper.frame.InvisibleMaxButton:SetShown(false);

MazeHelper.frame.MainHolder = CreateFrame('Frame', nil, MazeHelper.frame);
MazeHelper.frame.MainHolder:SetAllPoints();

-- Solution Text
MazeHelper.frame.SolutionText = MazeHelper.frame.MainHolder:CreateFontString(nil, 'ARTWORK', 'GameFontNormalLarge');
PixelUtil.SetPoint(MazeHelper.frame.SolutionText, 'LEFT', MazeHelper.frame, 'LEFT', 2, -54);
PixelUtil.SetPoint(MazeHelper.frame.SolutionText, 'RIGHT', MazeHelper.frame, 'RIGHT', -2, 0);
MazeHelper.frame.SolutionText:SetShadowColor(0.15, 0.15, 0.15);
MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_4']);

local function ResetAll()
    for i = 1, MAX_BUTTONS do
        MazeHelper:SetUnactiveButton(buttons[i]);

        buttons[i].state = false;
        buttons[i].sender = nil;
    end

    NUM_ACTIVE_BUTTONS = 0;
    SOLUTION_BUTTON_ID = nil;

    MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_4']);
    MazeHelper.frame.PassedButton:SetEnabled(false);
    MazeHelper.frame.AnnounceButton:SetShown(false);
    MazeHelper.frame.AnnounceButton.solutionAnnounceName = nil;

    MazeHelper.frame.MiniSolution:SetShown(false);
    MazeHelper.frame.PassedCounter:SetShown(true);

    MazeHelper.frame.ResetButton:SetEnabled(false);
end

MazeHelper.frame.ResetAll = ResetAll;

MazeHelper.frame.BottomButtonsHolder = CreateFrame('Frame', nil, MazeHelper.frame.MainHolder);
PixelUtil.SetPoint(MazeHelper.frame.BottomButtonsHolder, 'TOP', MazeHelper.frame.SolutionText, 'BOTTOM', 0, -8);
PixelUtil.SetSize(MazeHelper.frame.BottomButtonsHolder, MazeHelper.frame.MainHolder:GetWidth(), 22);

-- Reset Button
MazeHelper.frame.ResetButton = CreateFrame('Button', nil, MazeHelper.frame.BottomButtonsHolder, 'SharedButtonSmallTemplate');
PixelUtil.SetPoint(MazeHelper.frame.ResetButton, 'RIGHT', MazeHelper.frame.BottomButtonsHolder, 'RIGHT', 0, 0);
MazeHelper.frame.ResetButton:SetText(L['MAZE_HELPER_RESET']);
PixelUtil.SetSize(MazeHelper.frame.ResetButton, tonumber(MazeHelper.frame.ResetButton:GetTextWidth()) + 20, 22);
MazeHelper.frame.ResetButton:SetScript('OnClick', function()
    if NUM_ACTIVE_BUTTONS == 0 then
        return;
    end

    MazeHelper:SendResetCommand();
    ResetAll();
end);
MazeHelper.frame.ResetButton:SetEnabled(false);

-- Passed Button
MazeHelper.frame.PassedButton = CreateFrame('Button', nil, MazeHelper.frame.BottomButtonsHolder, 'SharedButtonSmallTemplate');
PixelUtil.SetPoint(MazeHelper.frame.PassedButton, 'RIGHT', MazeHelper.frame.ResetButton, 'LEFT', -8, 0);
MazeHelper.frame.PassedButton:SetText(L['MAZE_HELPER_PASSED']);
PixelUtil.SetSize(MazeHelper.frame.PassedButton, tonumber(MazeHelper.frame.PassedButton:GetTextWidth()) + 20, 22);
MazeHelper.frame.PassedButton:SetScript('OnClick', function()
    PASSED_COUNTER = PASSED_COUNTER + 1;

    MazeHelper.frame.PassedCounter.Text:SetText(PASSED_COUNTER);
    PixelUtil.SetPoint(MazeHelper.frame.PassedCounter.Text, 'CENTER', MazeHelper.frame.PassedCounter, 'CENTER', (PASSED_COUNTER == 1) and -2 or 0, isMinimized and 0 or -1);

    MazeHelper:SendPassedCommand(PASSED_COUNTER);

    ResetAll();
end);
MazeHelper.frame.PassedButton:SetEnabled(false);

PixelUtil.SetSize(MazeHelper.frame.BottomButtonsHolder, MazeHelper.frame.ResetButton:GetWidth() + MazeHelper.frame.PassedButton:GetWidth() + 8, 22);

-- Passed Counter Text
MazeHelper.frame.PassedCounter = CreateFrame('Frame', nil, MazeHelper.frame.MainHolder);
PixelUtil.SetPoint(MazeHelper.frame.PassedCounter, 'BOTTOM', MazeHelper.frame, 'TOP', 0, -32);
PixelUtil.SetSize(MazeHelper.frame.PassedCounter, 64, 64);
MazeHelper.frame.PassedCounter:SetScale(1.25);
MazeHelper.frame.PassedCounter.Background = MazeHelper.frame.PassedCounter:CreateTexture(nil, 'BACKGROUND');
MazeHelper.frame.PassedCounter.Background:SetAllPoints();
MazeHelper.frame.PassedCounter.Background:SetTexture(M.BBH);
MazeHelper.frame.PassedCounter.Text = MazeHelper.frame.PassedCounter:CreateFontString(nil, 'ARTWORK', 'GameFontNormalShadowHuge2');
PixelUtil.SetPoint(MazeHelper.frame.PassedCounter.Text, 'CENTER', MazeHelper.frame.PassedCounter, 'CENTER', -2, -1);
MazeHelper.frame.PassedCounter.Text:SetShadowColor(0.15, 0.15, 0.15);
MazeHelper.frame.PassedCounter.Text:SetText(PASSED_COUNTER);
MazeHelper.frame.PassedCounter.Text:SetJustifyH('CENTER');

-- Mini solution icon
MazeHelper.frame.MiniSolution = CreateFrame('Frame', nil, MazeHelper.frame.MainHolder);
MazeHelper.frame.MiniSolution:SetAllPoints(MazeHelper.frame.PassedCounter);
PixelUtil.SetPoint(MazeHelper.frame.MiniSolution, 'CENTER', MazeHelper.frame.PassedCounter, 'CENTER', 0, 0);
MazeHelper.frame.MiniSolution.Icon = MazeHelper.frame.MiniSolution:CreateTexture(nil, 'OVERLAY');
PixelUtil.SetPoint(MazeHelper.frame.MiniSolution.Icon, 'CENTER', MazeHelper.frame.MiniSolution, 'CENTER', 0, 0);
PixelUtil.SetSize(MazeHelper.frame.MiniSolution.Icon, 40, 40);
MazeHelper.frame.MiniSolution.Icon:SetTexture(M.Symbols.TEXTURE);
MazeHelper.frame.MiniSolution:SetShown(false);

-- Announce Button
MazeHelper.frame.AnnounceButton = CreateFrame('Button', nil, MazeHelper.frame.MainHolder);
PixelUtil.SetPoint(MazeHelper.frame.AnnounceButton, 'TOPLEFT', MazeHelper.frame, 'TOPLEFT', 2, 4);
PixelUtil.SetSize(MazeHelper.frame.AnnounceButton, 18, 18);
MazeHelper.frame.AnnounceButton:SetNormalTexture(M.Icons.TEXTURE);
MazeHelper.frame.AnnounceButton:GetNormalTexture():SetTexCoord(unpack(M.Icons.COORDS.MEGAPHONE_WHITE));
MazeHelper.frame.AnnounceButton:GetNormalTexture():SetVertexColor(0.7, 0.7, 0.7, 1);
MazeHelper.frame.AnnounceButton:SetHighlightTexture(M.Icons.TEXTURE, 'BLEND');
MazeHelper.frame.AnnounceButton:GetHighlightTexture():SetTexCoord(unpack(M.Icons.COORDS.MEGAPHONE_WHITE));
MazeHelper.frame.AnnounceButton:GetHighlightTexture():SetVertexColor(1, 0.85, 0, 1);
MazeHelper.frame.AnnounceButton:SetScript('OnClick', function(self)
    if not self.solutionAnnounceName then
        return;
    end

    local partyChatType = GetPartyChatType();
    if partyChatType then
        SendChatMessage(string.format(L['MAZE_HELPER_ANNOUNCE_SOLUTION'], self.solutionAnnounceName), partyChatType);
        self.solutionAnnounceName = nil;
    end

    self:SetShown(false);
end);
MazeHelper.frame.AnnounceButton:SetShown(false);

MazeHelper.frame.Settings = CreateFrame('Frame', nil, MazeHelper.frame);
MazeHelper.frame.Settings:SetAllPoints();
MazeHelper.frame.Settings:SetShown(false);

MazeHelper.frame.Settings.SyncEnabled = E.CreateRoundedCheckButton(MazeHelper.frame.Settings);
MazeHelper.frame.Settings.SyncEnabled:SetPosition('TOPLEFT', MazeHelper.frame.Settings, 'TOPLEFT', 12, -8);
MazeHelper.frame.Settings.SyncEnabled:SetArea(26, 26);
MazeHelper.frame.Settings.SyncEnabled:SetLabel(L['MAZE_HELPER_SETTINGS_SYNC_ENABLED_LABEL']);
MazeHelper.frame.Settings.SyncEnabled:SetTooltip(L['MAZE_HELPER_SETTINGS_SYNC_ENABLED_TOOLTIP']);
MazeHelper.frame.Settings.SyncEnabled:SetScript('OnClick', function(self)
    MHMOTSConfig.SyncEnabled = self:GetChecked();

    if MHMOTSConfig.SyncEnabled then
        MazeHelper.frame:RegisterEvent('CHAT_MSG_ADDON');
    else
        MazeHelper.frame:UnregisterEvent('CHAT_MSG_ADDON');
    end
end);

MazeHelper.frame.Settings.UseColoredSymbols = E.CreateRoundedCheckButton(MazeHelper.frame.Settings);
MazeHelper.frame.Settings.UseColoredSymbols:SetPosition('TOPLEFT', MazeHelper.frame.Settings.SyncEnabled, 'BOTTOMLEFT', 0, 0);
MazeHelper.frame.Settings.UseColoredSymbols:SetArea(26, 26);
MazeHelper.frame.Settings.UseColoredSymbols:SetLabel(L['MAZE_HELPER_SETTINGS_USE_COLORED_SYMBOLS_LABEL']);
MazeHelper.frame.Settings.UseColoredSymbols:SetTooltip(L['MAZE_HELPER_SETTINGS_USE_COLORED_SYMBOLS_TOOLTIP']);
MazeHelper.frame.Settings.UseColoredSymbols:SetScript('OnClick', function(self)
    MHMOTSConfig.UseColoredSymbols = self:GetChecked();
    SwitchSymbolsColorMode(MHMOTSConfig.UseColoredSymbols);
end);

MazeHelper.frame.Settings.PrintResettedPlayerName = E.CreateRoundedCheckButton(MazeHelper.frame.Settings);
MazeHelper.frame.Settings.PrintResettedPlayerName:SetPosition('TOPLEFT', MazeHelper.frame.Settings.UseColoredSymbols, 'BOTTOMLEFT', 0, 0);
MazeHelper.frame.Settings.PrintResettedPlayerName:SetArea(26, 26);
MazeHelper.frame.Settings.PrintResettedPlayerName:SetLabel(L['MAZE_HELPER_SETTINGS_REVEAL_RESETTER_LABEL']);
MazeHelper.frame.Settings.PrintResettedPlayerName:SetTooltip(L['MAZE_HELPER_SETTINGS_REVEAL_RESETTER_TOOLTIP']);
MazeHelper.frame.Settings.PrintResettedPlayerName:SetScript('OnClick', function(self)
    MHMOTSConfig.PrintResettedPlayerName = self:GetChecked();
end);

MazeHelper.frame.Settings.ShowAtBoss = E.CreateRoundedCheckButton(MazeHelper.frame.Settings);
MazeHelper.frame.Settings.ShowAtBoss:SetPosition('TOPLEFT', MazeHelper.frame.Settings.PrintResettedPlayerName, 'BOTTOMLEFT', 0, 0);
MazeHelper.frame.Settings.ShowAtBoss:SetArea(26, 26);
MazeHelper.frame.Settings.ShowAtBoss:SetLabel(L['MAZE_HELPER_SETTINGS_SHOW_AT_BOSS_LABEL']);
MazeHelper.frame.Settings.ShowAtBoss:SetTooltip(L['MAZE_HELPER_SETTINGS_SHOW_AT_BOSS_TOOLTIP']);
MazeHelper.frame.Settings.ShowAtBoss:SetScript('OnClick', function(self)
    MHMOTSConfig.ShowAtBoss = self:GetChecked();
end);

MazeHelper.frame.Settings.StartInMinMode = E.CreateRoundedCheckButton(MazeHelper.frame.Settings);
MazeHelper.frame.Settings.StartInMinMode:SetPosition('TOPLEFT', MazeHelper.frame.Settings.ShowAtBoss, 'BOTTOMLEFT', 0, 0);
MazeHelper.frame.Settings.StartInMinMode:SetArea(26, 26);
MazeHelper.frame.Settings.StartInMinMode:SetLabel(L['MAZE_HELPER_SETTINGS_START_IN_MINMODE_LABEL']);
MazeHelper.frame.Settings.StartInMinMode:SetTooltip(L['MAZE_HELPER_SETTINGS_START_IN_MINMODE_TOOLTIP']);
MazeHelper.frame.Settings.StartInMinMode:SetScript('OnClick', function(self)
    MHMOTSConfig.StartInMinMode = self:GetChecked();
end);

MazeHelper.frame.Settings.WASupport = E.CreateRoundedCheckButton(MazeHelper.frame.Settings);
MazeHelper.frame.Settings.WASupport:SetPosition('TOPLEFT', MazeHelper.frame.Settings.StartInMinMode, 'BOTTOMLEFT', 0, 0);
MazeHelper.frame.Settings.WASupport:SetArea(26, 26);
MazeHelper.frame.Settings.WASupport:SetLabel(L['MAZE_HELPER_SETTINGS_WA_SUPPORT_LABEL']);
MazeHelper.frame.Settings.WASupport:SetTooltip(L['MAZE_HELPER_SETTINGS_WA_SUPPORT_TOOLTIP']);
MazeHelper.frame.Settings.WASupport:SetScript('OnClick', function(self)
    MHMOTSConfig.WASupport = self:GetChecked();

    if MHMOTSConfig.WASupport then
        MazeHelper.frame:RegisterEvent('CHAT_MSG_ADDON_LOGGED');
    else
        MazeHelper.frame:UnregisterEvent('CHAT_MSG_ADDON_LOGGED');
    end
end);

MazeHelper.frame.Settings.AutoAnnouncer = E.CreateRoundedCheckButton(MazeHelper.frame.Settings);
MazeHelper.frame.Settings.AutoAnnouncer:SetPosition('TOPLEFT', MazeHelper.frame.Settings.WASupport, 'BOTTOMLEFT', 0, 0);
MazeHelper.frame.Settings.AutoAnnouncer:SetArea(26, 26);
MazeHelper.frame.Settings.AutoAnnouncer:SetLabel(L['MAZE_HELPER_SETTINGS_AUTOANNOUNCER_LABEL']);
MazeHelper.frame.Settings.AutoAnnouncer:SetTooltip(L['MAZE_HELPER_SETTINGS_AUTOANNOUNCER_TOOLTIP']);
MazeHelper.frame.Settings.AutoAnnouncer:SetScript('OnClick', function(self)
    MHMOTSConfig.AutoAnnouncer = self:GetChecked();

    MazeHelper.frame.Settings.AutoAnnouncerAsPartyLeader:SetEnabled(MHMOTSConfig.AutoAnnouncer);
    MazeHelper.frame.Settings.AutoAnnouncerAsAlways:SetEnabled(MHMOTSConfig.AutoAnnouncer);
    MazeHelper.frame.Settings.AutoAnnouncerAsTank:SetEnabled(MHMOTSConfig.AutoAnnouncer);
    MazeHelper.frame.Settings.AutoAnnouncerAsHealer:SetEnabled(MHMOTSConfig.AutoAnnouncer);
end);

MazeHelper.frame.Settings.AutoAnnouncerAsPartyLeader = E.CreateCheckButton('MazeHelper_Settings_AutoAnnouncerAsPartyLeader_CheckButton', MazeHelper.frame.Settings);
MazeHelper.frame.Settings.AutoAnnouncerAsPartyLeader:SetPosition('TOPLEFT', MazeHelper.frame.Settings.AutoAnnouncer, 'BOTTOMRIGHT', 0, 2);
PixelUtil.SetPoint(MazeHelper.frame.Settings.AutoAnnouncerAsPartyLeader.Label, 'LEFT', MazeHelper.frame.Settings.AutoAnnouncerAsPartyLeader, 'RIGHT', 4, 2);
MazeHelper.frame.Settings.AutoAnnouncerAsPartyLeader:SetLabel(M.INLINE_LEADER_ICON);
MazeHelper.frame.Settings.AutoAnnouncerAsPartyLeader:SetTooltip(L['MAZE_HELPER_SETTINGS_AA_PARTY_LEADER']);
MazeHelper.frame.Settings.AutoAnnouncerAsPartyLeader:SetScript('OnClick', function(self)
    MHMOTSConfig.AutoAnnouncerAsPartyLeader = self:GetChecked();
end);

MazeHelper.frame.Settings.AutoAnnouncerAsAlways = E.CreateCheckButton('MazeHelper_Settings_AutoAnnouncerAsAlways_CheckButton', MazeHelper.frame.Settings);
MazeHelper.frame.Settings.AutoAnnouncerAsAlways:SetPosition('LEFT', MazeHelper.frame.Settings.AutoAnnouncerAsPartyLeader.Label, 'RIGHT', 12, -2);
MazeHelper.frame.Settings.AutoAnnouncerAsAlways:SetLabel(M.INLINE_INFINITY_ICON);
MazeHelper.frame.Settings.AutoAnnouncerAsAlways:SetTooltip(L['MAZE_HELPER_SETTINGS_AA_ALWAYS']);
MazeHelper.frame.Settings.AutoAnnouncerAsAlways:SetScript('OnClick', function(self)
    MHMOTSConfig.AutoAnnouncerAsAlways = self:GetChecked();
end);

MazeHelper.frame.Settings.AutoAnnouncerAsTank = E.CreateCheckButton('MazeHelper_Settings_AutoAnnouncerAsTank_CheckButton', MazeHelper.frame.Settings);
MazeHelper.frame.Settings.AutoAnnouncerAsTank:SetPosition('LEFT', MazeHelper.frame.Settings.AutoAnnouncerAsAlways.Label, 'RIGHT', 12, 0);
MazeHelper.frame.Settings.AutoAnnouncerAsTank:SetLabel(M.INLINE_TANK_ICON);
MazeHelper.frame.Settings.AutoAnnouncerAsTank:SetTooltip(L['MAZE_HELPER_SETTINGS_AA_TANK']);
MazeHelper.frame.Settings.AutoAnnouncerAsTank:SetScript('OnClick', function(self)
    MHMOTSConfig.AutoAnnouncerAsTank = self:GetChecked();
end);

MazeHelper.frame.Settings.AutoAnnouncerAsHealer = E.CreateCheckButton('MazeHelper_Settings_AutoAnnouncerAsHealer_CheckButton', MazeHelper.frame.Settings);
MazeHelper.frame.Settings.AutoAnnouncerAsHealer:SetPosition('LEFT', MazeHelper.frame.Settings.AutoAnnouncerAsTank.Label, 'RIGHT', 12, 0);
MazeHelper.frame.Settings.AutoAnnouncerAsHealer:SetLabel(M.INLINE_HEALER_ICON);
MazeHelper.frame.Settings.AutoAnnouncerAsHealer:SetTooltip(L['MAZE_HELPER_SETTINGS_AA_HEALER']);
MazeHelper.frame.Settings.AutoAnnouncerAsHealer:SetScript('OnClick', function(self)
    MHMOTSConfig.AutoAnnouncerAsHealer = self:GetChecked();
end);

function MazeHelper:CreateButton(index)
    local button = CreateFrame('Button', nil, MazeHelper.frame.MainHolder, 'BackdropTemplate');

    if index == 1 then
        PixelUtil.SetPoint(button, 'TOPLEFT', MazeHelper.frame.MainHolder, 'TOPLEFT', 20, -20);
    elseif index == 5 then
        PixelUtil.SetPoint(button, 'TOPLEFT', buttons[1], 'BOTTOMLEFT', 0, Y_OFFSET);
    else
        PixelUtil.SetPoint(button, 'LEFT', buttons[index - 1], 'RIGHT', X_OFFSET, 0);
    end

    PixelUtil.SetSize(button, 64, 64);

    button.Icon = button:CreateTexture(nil, 'ARTWORK');
    PixelUtil.SetPoint(button.Icon, 'TOPLEFT', button, 'TOPLEFT', 4, -4);
    PixelUtil.SetPoint(button.Icon, 'BOTTOMRIGHT', button, 'BOTTOMRIGHT', -4, 4);
    button.Icon:SetTexture(M.Symbols.TEXTURE);
    button.Icon:SetTexCoord(unpack(MHMOTSConfig.UseColoredSymbols and buttonsData[index].coords or buttonsData[index].coords_white));

    button:SetBackdrop({
        insets = {top = 1, left = 1, bottom = 1, right = 1},
        edgeFile = 'Interface\\Buttons\\WHITE8x8',
		edgeSize = 2,
    });

    MazeHelper:SetUnactiveButton(button);

    button.id    = index;
    button.data  = buttonsData[index];
    button.state = false;

    button:RegisterForClicks('LeftButtonUp', 'RightButtonUp');

    button:SetScript('OnClick', function(self, b)
        if b == 'LeftButton' then
            if self.state or NUM_ACTIVE_BUTTONS == MAX_ACTIVE_BUTTONS then
                return;
            end

            MazeHelper.frame.ResetButton:SetEnabled(true);

            NUM_ACTIVE_BUTTONS = math.min(MAX_ACTIVE_BUTTONS, NUM_ACTIVE_BUTTONS + 1);
            self.state = true;
            MazeHelper:SetActiveButton(self);
            MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_' .. (MAX_ACTIVE_BUTTONS - NUM_ACTIVE_BUTTONS)]);

            MazeHelper:SendButtonID(self.id, 'ACTIVE');

            if NUM_ACTIVE_BUTTONS == MAX_ACTIVE_BUTTONS then
                MazeHelper:UpdateSolution();
            end
        elseif b == 'RightButton' then
            if not self.state then
                return;
            end

            if self.state then
                NUM_ACTIVE_BUTTONS = math.max(0, NUM_ACTIVE_BUTTONS - 1);
                self.state = false;
                self.sender = nil;
            end

            if NUM_ACTIVE_BUTTONS < MAX_ACTIVE_BUTTONS then
                MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_' .. (MAX_ACTIVE_BUTTONS - NUM_ACTIVE_BUTTONS)]);

                for i = 1, MAX_BUTTONS do
                    if buttons[i].state then
                        MazeHelper:SetActiveButton(buttons[i]);
                    end
                end

                MazeHelper.frame.PassedButton:SetEnabled(false);
                MazeHelper.frame.AnnounceButton:SetShown(false);
                MazeHelper.frame.AnnounceButton.solutionAnnounceName = nil;

                SOLUTION_BUTTON_ID = nil;
            end

            if NUM_ACTIVE_BUTTONS == 0 then
                MazeHelper.frame.ResetButton:SetEnabled(false);
            end

            MazeHelper:SetUnactiveButton(self);
            MazeHelper:SendButtonID(self.id, 'UNACTIVE');
        end
    end);

    button:SetScript('OnEnter', function(self)
        if not self.sender then
            return;
        end

        GameTooltip:SetOwner(self, 'ANCHOR_RIGHT');
        GameTooltip:AddLine(self.state and string.format(L['MAZE_HELPER_SENDED_BY'], self.sender) or string.format(L['MAZE_HELPER_CLEARED_BY'], self.sender), 1, 0.85, 0, true);
        GameTooltip:Show();
    end);
    button:SetScript('OnLeave', GameTooltip_Hide);

    button:RegisterForDrag('LeftButton');
    button:SetScript('OnDragStart', function()
        if MazeHelper.frame:IsMovable() then
            MazeHelper.frame:StartMoving();
        end
    end);
    button:SetScript('OnDragStop', function()
        BetterOnDragStop(MazeHelper.frame);
    end);

    table.insert(buttons, index, button); -- index for just to be sure
end

function MazeHelper:CreateButtons()
    for i = 1, MAX_BUTTONS do
        MazeHelper:CreateButton(i);
    end
end

function MazeHelper:SetUnactiveButton(button)
    button:SetBackdropBorderColor(0, 0, 0, 0);
end

function MazeHelper:SetActiveButton(button)
    button:SetBackdropBorderColor(0.4, 0.52, 0.95, 1);
end

function MazeHelper:SetReceivedButton(button)
    button:SetBackdropBorderColor(0.9, 1, 0.1, 1);
end

function MazeHelper:SetSolutionButton(button)
    button:SetBackdropBorderColor(0.2, 0.8, 0.4, 1);
end

-- DON'T LOOK AT THIS SHIT PLEEEEASE :( I WAS DRUNK, BUT IT STILL WORKS SOMEHOW
function MazeHelper:UpdateSolution()
    local circleSum, flowerSum, leafSum, fillSum = 0, 0, 0, 0;
    for i = 1, MAX_BUTTONS do
        if buttons[i].state then
            if buttons[i].data.circle then
                circleSum = circleSum + 1;
            end

            if buttons[i].data.flower then
                flowerSum = flowerSum + 1;
            end

            if buttons[i].data.leaf then
                leafSum = leafSum + 1;
            end

            if buttons[i].data.fill then
                fillSum = fillSum + 1;
            end
        end
    end

    local fill, flower, leaf, circle;
    if fillSum == 3 then
        fill = false;
    elseif fillSum == 1 then
        fill = true;
    end

    if flowerSum == 3 then
        flower = false
    elseif flowerSum == 1 then
        flower = true;
    end

    if leafSum == 3 then
        leaf = false;
    elseif leafSum == 1 then
        leaf = true;
    end

    if circleSum == 3 then
        circle = false;
    elseif circleSum == 1 then
        circle = true;
    end

    SOLUTION_BUTTON_ID = nil;

    local solutionName, solutionAnnounceName;
    for i = 1, MAX_BUTTONS do
        if buttons[i].state then
            if buttons[i].data.fill == fill then
                SOLUTION_BUTTON_ID = i;
            end

            if buttons[i].data.leaf == leaf then
                SOLUTION_BUTTON_ID = i;
            end

            if buttons[i].data.flower == flower then
                SOLUTION_BUTTON_ID = i;
            end

            if buttons[i].data.circle == circle then
                SOLUTION_BUTTON_ID = i;
            end
        end
    end

    solutionName         = SOLUTION_BUTTON_ID and buttons[SOLUTION_BUTTON_ID].data.name;
    solutionAnnounceName = SOLUTION_BUTTON_ID and buttons[SOLUTION_BUTTON_ID].data.aname;

    if SOLUTION_BUTTON_ID then
        MazeHelper:SetSolutionButton(buttons[SOLUTION_BUTTON_ID]);
        MazeHelper.frame.MiniSolution.Icon:SetTexCoord(unpack(MHMOTSConfig.UseColoredSymbols and buttonsData[SOLUTION_BUTTON_ID].coords or buttonsData[SOLUTION_BUTTON_ID].coords_white));

        if isMinimized then
            MazeHelper.frame.MiniSolution:SetShown(true);
            MazeHelper.frame.PassedCounter:SetShown(false);
        end
    else
        MazeHelper.frame.MiniSolution:SetShown(false);
        MazeHelper.frame.PassedCounter:SetShown(true);
    end

    local partyChatType = GetPartyChatType();

    MazeHelper.frame.AnnounceButton.solutionAnnounceName = solutionAnnounceName or nil;
    MazeHelper.frame.AnnounceButton:SetShown((not isMinimized and solutionAnnounceName and partyChatType and not MHMOTSConfig.AutoAnnouncer) and true or false);

    if MHMOTSConfig.AutoAnnouncer and solutionAnnounceName and partyChatType then
        local announce = false;

        if MHMOTSConfig.AutoAnnouncerAsAlways then
            announce = true;
        elseif MHMOTSConfig.AutoAnnouncerAsPartyLeader and UnitIsGroupLeader('player') then
            announce = true;
        elseif MHMOTSConfig.AutoAnnouncerAsTank and playerRole == 'TANK' then
            announce = true;
        elseif MHMOTSConfig.AutoAnnouncerAsHealer and playerRole == 'HEALER' then
            announce = true;
        end

        if announce then
            SendChatMessage(string.format(L['MAZE_HELPER_ANNOUNCE_SOLUTION'], solutionAnnounceName), partyChatType);
        end
    end

    MazeHelper.frame.PassedButton:SetEnabled(SOLUTION_BUTTON_ID and true or false);
    MazeHelper.frame.SolutionText:SetText(solutionName and string.format(L['MAZE_HELPER_SOLUTION'], solutionName) or L['MAZE_HELPER_SOLUTION_NA']);
end

function MazeHelper:SendResetCommand()
    local partyChatType = GetPartyChatType();
    if partyChatType then
        if MHMOTSConfig.SyncEnabled then
            C_ChatInfo.SendAddonMessage('MAZEHELPER', 'SendReset', partyChatType);
        end

        -- Mists of Tirna Scithe Helper addon compatibility
        C_ChatInfo.SendAddonMessage('MOTSH', 'reset', partyChatType);

        -- WeakAura (https://wago.io/hVLym_eLv) compatibility
        if MHMOTSConfig.WASupport then
            C_ChatInfo.SendAddonMessageLogged('TIRNASCYTHE', '9 9 9 9 9', partyChatType);
        end
    end
end

function MazeHelper:SendPassedCommand(step)
    local partyChatType = GetPartyChatType();
    if partyChatType then
        if MHMOTSConfig.SyncEnabled then
            local data = 'SendPassed|'.. step;
            C_ChatInfo.SendAddonMessage('MAZEHELPER', data, partyChatType);
        end

        -- Mists of Tirna Scithe Helper addon compatibility
        C_ChatInfo.SendAddonMessage('MOTSH', 'reset', partyChatType);

        -- WeakAura (https://wago.io/hVLym_eLv) compatibility
        if MHMOTSConfig.WASupport then
            C_ChatInfo.SendAddonMessageLogged('TIRNASCYTHE', '9 9 9 9 9', partyChatType);
        end
    end
end

function MazeHelper:SendPassedCounter()
    if not MHMOTSConfig.SyncEnabled then
        return;
    end

    local partyChatType = GetPartyChatType();
    if not partyChatType then
        return;
    end

    C_ChatInfo.SendAddonMessage('MAZEHELPER', 'RECPC|' .. PASSED_COUNTER, partyChatType);
end

function MazeHelper:RequestPassedCounter()
    if not MHMOTSConfig.SyncEnabled then
        return;
    end

    local partyChatType = GetPartyChatType();
    if not partyChatType then
        return;
    end

    C_ChatInfo.SendAddonMessage('MAZEHELPER', 'REQPC', partyChatType);
end

function MazeHelper:SendButtonID(buttonID, mode)
    local partyChatType = GetPartyChatType();
    if partyChatType then
        if MHMOTSConfig.SyncEnabled then
            local data = 'SendButtonID|' .. buttonID .. '|' .. mode;
            C_ChatInfo.SendAddonMessage('MAZEHELPER', data, partyChatType);
        end

        -- Mists of Tirna Scithe Helper addon compatibility
        C_ChatInfo.SendAddonMessage('MOTSH', buttonsMOTSHMap[buttonID], partyChatType);

        -- WeakAura (https://wago.io/hVLym_eLv) compatibility
        if MHMOTSConfig.WASupport then
            if mode ~= 'UNACTIVE' then
                C_ChatInfo.SendAddonMessageLogged('TIRNASCYTHE', buttonsMYtoWA[buttonID] .. ' 9', partyChatType);
            end
        end
    end
end

function MazeHelper:ReceiveResetCommand()
    ResetAll();
end

function MazeHelper:ReceivePassedCommand(step)
    PASSED_COUNTER = step;

    MazeHelper.frame.PassedCounter.Text:SetText(PASSED_COUNTER);
    PixelUtil.SetPoint(MazeHelper.frame.PassedCounter.Text, 'CENTER', MazeHelper.frame.PassedCounter, 'CENTER', (PASSED_COUNTER == 1) and -2 or 0, isMinimized and 0 or -1);
    ResetAll();
end

function MazeHelper:ReceivePassedCounter(step)
    if step and step == PASSED_COUNTER then
        return;
    end

    PASSED_COUNTER = step;

    MazeHelper.frame.PassedCounter.Text:SetText(PASSED_COUNTER);
    PixelUtil.SetPoint(MazeHelper.frame.PassedCounter.Text, 'CENTER', MazeHelper.frame.PassedCounter, 'CENTER', (PASSED_COUNTER == 1) and -2 or 0, isMinimized and 0 or -1);
end

function MazeHelper:ReceiveActiveButtonID(buttonID, sender)
    if not buttons[buttonID] then
        return;
    end

    if buttons[buttonID].state or NUM_ACTIVE_BUTTONS == MAX_ACTIVE_BUTTONS then
        return;
    end

    NUM_ACTIVE_BUTTONS = math.min(MAX_ACTIVE_BUTTONS, NUM_ACTIVE_BUTTONS + 1);
    buttons[buttonID].state = true;
    buttons[buttonID].sender = sender;

    MazeHelper:SetReceivedButton(buttons[buttonID]);
    MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_' .. (MAX_ACTIVE_BUTTONS - NUM_ACTIVE_BUTTONS)]);

    MazeHelper.frame.ResetButton:SetEnabled(true);

    if NUM_ACTIVE_BUTTONS == MAX_ACTIVE_BUTTONS then
        MazeHelper:UpdateSolution();
    end
end

function MazeHelper:ReceiveUnactiveButtonID(buttonID, sender)
    if not buttons[buttonID] then
        return;
    end

    if not buttons[buttonID].state then
        return;
    end

    NUM_ACTIVE_BUTTONS = math.max(0, NUM_ACTIVE_BUTTONS - 1);
    buttons[buttonID].state = false;
    buttons[buttonID].sender = sender;

    if NUM_ACTIVE_BUTTONS < MAX_ACTIVE_BUTTONS then
        MazeHelper.frame.SolutionText:SetText(L['MAZE_HELPER_CHOOSE_SYMBOLS_' .. (MAX_ACTIVE_BUTTONS - NUM_ACTIVE_BUTTONS)]);

        for i = 1, MAX_BUTTONS do
            if buttons[i].state then
                MazeHelper:SetActiveButton(buttons[i]);
            end
        end

        MazeHelper.frame.PassedButton:SetEnabled(false);
        MazeHelper.frame.AnnounceButton:SetShown(false);
        MazeHelper.frame.AnnounceButton.solutionAnnounceName = nil;

        SOLUTION_BUTTON_ID = nil;
    end

    if NUM_ACTIVE_BUTTONS == 0 then
        MazeHelper.frame.ResetButton:SetEnabled(false);
    end

    MazeHelper:SetUnactiveButton(buttons[buttonID]);
end

local function UpdateShown()
    if MHMOTSConfig.ShowAtBoss then
        MazeHelper.frame:SetShown((not bossKilled and inInstance and GetMinimapZoneText() == L['MAZE_HELPER_ZONE_NAME']));
    else
        MazeHelper.frame:SetShown((not inEncounter and inInstance and GetMinimapZoneText() == L['MAZE_HELPER_ZONE_NAME']));
    end

    if MazeHelper.frame:IsShown() then
        if MHMOTSConfig.StartInMinMode and not startedInMinMode then
            MazeHelper.frame.MinButton:Click();
            startedInMinMode = true;
        end
    end

    if inEncounter then
        PixelUtil.SetSize(MazeHelper.frame.BottomButtonsHolder, MazeHelper.frame.ResetButton:GetWidth(), 22);

        MazeHelper.frame.PassedButton:Hide();
        MazeHelper.frame.PassedCounter:Hide();
    else
        PixelUtil.SetSize(MazeHelper.frame.BottomButtonsHolder, MazeHelper.frame.ResetButton:GetWidth() + MazeHelper.frame.PassedButton:GetWidth() + 8, 22);

        MazeHelper.frame.PassedButton:Show();
        MazeHelper.frame.PassedCounter:Show();
    end
end

MazeHelper.frame.UpdateShown = UpdateShown;

MazeHelper.frame:RegisterEvent('ADDON_LOADED');
MazeHelper.frame:SetScript('OnEvent', function(self, event, ...)
    if self[event] then
        return self[event](self, ...);
    end
end);

local function UpdateData(frame) -- Not good name, i know...
    -- Sometimes on PLAYER_LOGIN GetNormalizedRealmName() don't return realm name, so we repeat it at PLAYER_ENTERING_WORLD
    playerNameWithRealm = UnitName('player') .. '-' .. (GetNormalizedRealmName() or string.gsub(GetRealmName(), '[%s%-]', '')); -- Just for sure...
    inInstance = IsInInstance();
    bossKilled = inInstance and (select(3, GetInstanceLockTimeRemainingEncounter(2))) or false;
    inEncounter = not bossKilled and UnitExists('boss1');

    PASSED_COUNTER = 1;
    MazeHelper.frame.PassedCounter.Text:SetText(PASSED_COUNTER);
    PixelUtil.SetPoint(MazeHelper.frame.PassedCounter.Text, 'CENTER', MazeHelper.frame.PassedCounter, 'CENTER', (PASSED_COUNTER == 1) and -2 or 0, isMinimized and 0 or -1);

    MazeHelper:RequestPassedCounter(); -- if you were dc'ed or reloading ui

    startedInMinMode = false;

    if inInstance then
        frame:RegisterEvent('ZONE_CHANGED');
        frame:RegisterEvent('ZONE_CHANGED_INDOORS');
        frame:RegisterEvent('ZONE_CHANGED_NEW_AREA');
        frame:RegisterEvent('ENCOUNTER_START');
        frame:RegisterEvent('ENCOUNTER_END');
        frame:RegisterEvent('BOSS_KILL');
    else
        frame:UnregisterEvent('ZONE_CHANGED');
        frame:UnregisterEvent('ZONE_CHANGED_INDOORS');
        frame:UnregisterEvent('ZONE_CHANGED_NEW_AREA');
        frame:UnregisterEvent('ENCOUNTER_START');
        frame:UnregisterEvent('ENCOUNTER_END');
        frame:UnregisterEvent('BOSS_KILL');
    end

    UpdateShown();
end

function MazeHelper.frame:PLAYER_LOGIN()
    UpdateData(self);
end

function MazeHelper.frame:PLAYER_ENTERING_WORLD()
    UpdateData(self);
end

function MazeHelper.frame:ZONE_CHANGED()
    UpdateShown();
end

function MazeHelper.frame:ZONE_CHANGED_INDOORS()
    UpdateShown();
end

function MazeHelper.frame:ZONE_CHANGED_NEW_AREA()
    UpdateShown();
end

local function UpdateBossState(encounterID, inFight, killed)
    if encounterID ~= MISTCALLER_ENCOUNTER_ID then
        return;
    end

    inEncounter = inFight;

    ResetAll();

    local partyChatType = GetPartyChatType();
    if partyChatType then
        -- Mists of Tirna Scithe Helper addon compatibility
        C_ChatInfo.SendAddonMessage('MOTSH', 'reset', partyChatType);

        -- WeakAura (https://wago.io/hVLym_eLv) compatibility
        if MHMOTSConfig.WASupport then
            C_ChatInfo.SendAddonMessageLogged('TIRNASCYTHE', '9 9 9 9 9', partyChatType);
        end
    end

    bossKilled = killed;
    UpdateShown();
end

function MazeHelper.frame:ENCOUNTER_START(encounterID)
    UpdateBossState(encounterID, true, false);
end

function MazeHelper.frame:ENCOUNTER_END(encounterID, _, _, _, success)
    UpdateBossState(encounterID, false, success);
end

function MazeHelper.frame:BOSS_KILL(encounterID)
    UpdateBossState(encounterID, false, true);
end

function MazeHelper.frame:CHAT_MSG_ADDON(prefix, message, _, sender)
    if sender == playerNameWithRealm then
        return;
    end

    if prefix == 'MAZEHELPER' then
        local p, buttonID, mode = strsplit('|', message);
        if p == 'SendButtonID'  then
            if mode == 'ACTIVE' then
                MazeHelper:ReceiveActiveButtonID(tonumber(buttonID), sender);
            elseif mode == 'UNACTIVE' then
                MazeHelper:ReceiveUnactiveButtonID(tonumber(buttonID), sender);
            end
        elseif p == 'SendPassed' then
            MazeHelper:ReceivePassedCommand(tonumber(buttonID));

            if MHMOTSConfig.PrintResettedPlayerName then
                print(string.format(L['MAZE_HELPER_PASSED_PLAYER'], sender));
            end
        elseif p == 'SendReset' then
            MazeHelper:ReceiveResetCommand();

            if MHMOTSConfig.PrintResettedPlayerName then
                print(string.format(L['MAZE_HELPER_RESETED_PLAYER'], sender));
            end
        elseif p == 'REQPC' then
            MazeHelper:SendPassedCounter();
        elseif p == 'RECPC' then
            MazeHelper:ReceivePassedCounter(tonumber(buttonID));
        end
	end
end

-- WHY?! JUST WHY?!
function MazeHelper.frame:CHAT_MSG_ADDON_LOGGED(prefix, message, _, sender)
    if sender == playerNameWithRealm then
        return;
    end

    if prefix == 'TIRNASCYTHE' then
        if message == '9 9 9 9' then
            MazeHelper:ReceiveResetCommand();

            if MHMOTSConfig.PrintResettedPlayerName then
                print(string.format(L['MAZE_HELPER_WA_RESETED_PLAYER'], sender));
            end
        else
            if buttonsWAtoMY[message] then
                MazeHelper:ReceiveActiveButtonID(buttonsWAtoMY[message], sender);
            end
        end
    end
end

function MazeHelper.frame:PLAYER_SPECIALIZATION_CHANGED(unit)
    if unit ~= 'player' then
        return;
    end

    playerRole = (select(5, GetSpecializationInfo(GetSpecialization())) or '');
end

function MazeHelper.frame:ADDON_LOADED(addonName)
    if addonName ~= ADDON_NAME then
        return;
    end

    self:UnregisterEvent('ADDON_LOADED');

    MHMOTSConfig = MHMOTSConfig or {};
    MHMOTSConfig.SyncEnabled             = MHMOTSConfig.SyncEnabled == nil and true or MHMOTSConfig.SyncEnabled;
    MHMOTSConfig.PrintResettedPlayerName = MHMOTSConfig.PrintResettedPlayerName == nil and true or MHMOTSConfig.PrintResettedPlayerName;
    MHMOTSConfig.ShowAtBoss              = MHMOTSConfig.ShowAtBoss == nil and true or MHMOTSConfig.ShowAtBoss;
    MHMOTSConfig.StartInMinMode          = MHMOTSConfig.StartInMinMode == nil and false or MHMOTSConfig.StartInMinMode;
    MHMOTSConfig.UseColoredSymbols       = MHMOTSConfig.UseColoredSymbols == nil and true or MHMOTSConfig.UseColoredSymbols;
    MHMOTSConfig.WASupport               = MHMOTSConfig.WASupport == nil and false or MHMOTSConfig.WASupport;

    MHMOTSConfig.AutoAnnouncer              = MHMOTSConfig.AutoAnnouncer == nil and false or MHMOTSConfig.AutoAnnouncer;
    MHMOTSConfig.AutoAnnouncerAsPartyLeader = MHMOTSConfig.AutoAnnouncerAsPartyLeader == nil and true or MHMOTSConfig.AutoAnnouncerAsPartyLeader;
    MHMOTSConfig.AutoAnnouncerAsAlways      = MHMOTSConfig.AutoAnnouncerAsAlways == nil and false or MHMOTSConfig.AutoAnnouncerAsAlways;
    MHMOTSConfig.AutoAnnouncerAsTank        = MHMOTSConfig.AutoAnnouncerAsTank == nil and false or MHMOTSConfig.AutoAnnouncerAsTank;
    MHMOTSConfig.AutoAnnouncerAsHealer      = MHMOTSConfig.AutoAnnouncerAsHealer == nil and false or MHMOTSConfig.AutoAnnouncerAsHealer;

    MazeHelper.frame.Settings.SyncEnabled:SetChecked(MHMOTSConfig.SyncEnabled);
    MazeHelper.frame.Settings.UseColoredSymbols:SetChecked(MHMOTSConfig.UseColoredSymbols);
    MazeHelper.frame.Settings.PrintResettedPlayerName:SetChecked(MHMOTSConfig.PrintResettedPlayerName);
    MazeHelper.frame.Settings.ShowAtBoss:SetChecked(MHMOTSConfig.ShowAtBoss);
    MazeHelper.frame.Settings.StartInMinMode:SetChecked(MHMOTSConfig.StartInMinMode);
    MazeHelper.frame.Settings.WASupport:SetChecked(MHMOTSConfig.WASupport);

    MazeHelper.frame.Settings.AutoAnnouncer:SetChecked(MHMOTSConfig.AutoAnnouncer);
    MazeHelper.frame.Settings.AutoAnnouncerAsPartyLeader:SetChecked(MHMOTSConfig.AutoAnnouncerAsPartyLeader);
    MazeHelper.frame.Settings.AutoAnnouncerAsAlways:SetChecked(MHMOTSConfig.AutoAnnouncerAsAlways);
    MazeHelper.frame.Settings.AutoAnnouncerAsTank:SetChecked(MHMOTSConfig.AutoAnnouncerAsTank);
    MazeHelper.frame.Settings.AutoAnnouncerAsHealer:SetChecked(MHMOTSConfig.AutoAnnouncerAsHealer);

    MazeHelper.frame.Settings.AutoAnnouncerAsPartyLeader:SetEnabled(MHMOTSConfig.AutoAnnouncer);
    MazeHelper.frame.Settings.AutoAnnouncerAsAlways:SetEnabled(MHMOTSConfig.AutoAnnouncer);
    MazeHelper.frame.Settings.AutoAnnouncerAsTank:SetEnabled(MHMOTSConfig.AutoAnnouncer);
    MazeHelper.frame.Settings.AutoAnnouncerAsHealer:SetEnabled(MHMOTSConfig.AutoAnnouncer);

    MazeHelper:CreateButtons();

    self:RegisterEvent('PLAYER_LOGIN');
    self:RegisterEvent('PLAYER_ENTERING_WORLD');
    self:RegisterEvent('PLAYER_SPECIALIZATION_CHANGED');

    if MHMOTSConfig.SyncEnabled then
        self:RegisterEvent('CHAT_MSG_ADDON');
    end

    if MHMOTSConfig.WASupport then
        self:RegisterEvent('CHAT_MSG_ADDON_LOGGED');
    end

    _G['SLASH_MAZEHELPER1'] = '/mh';
    SlashCmdList['MAZEHELPER'] = function()
        MazeHelper.frame:SetShown(not MazeHelper.frame:IsShown());
    end
end